@@ -175,51 +175,35 @@ bz_decompress (const char *zpath, char *dest)
     return false;
 }
 
-/* return type of protocol of url saved in `str' */
-protocol
-check_proto_t (const char *str)
-{
-    if (!strncasecmp(str, "http:", 5))
-	 return http;
-
-    if (!strncasecmp(str, "ftp:", 4))
-	 return ftp;
-
-    _error (3, "\n"
-    "Fatal error: repository \"", str,
-	 "\" isn't HTTP or FTP.\n"
-	 "Check your slackyd.conf.\n\n");
-
-    return 0; /* :-m */
-}
 
 /* make update data directories */
 static void
 update_dir (void)
 {
-	const char *dbranch[BRANCHES] = { "/", "/patches", "/extra" };
+	const char *cache = opt.cache;
+    const char *dbranch[BRANCHES] = { "/", "/patches", "/extra" };
 	unsigned r = 0;
 	char path[BUFFSIZE0];
 	bool slackware = false;
 	
-	for (r = 0; r < REPOS[0].N; r++) {
+	for (r = 0; r < N_REPOS; r++) {
 		
 	if (REPOS[r].SLACKWARE && slackware)
 	 continue; /* we use only one official mirror */
 		
 	/* repository directory */
-	snprintf (path, BUFFSIZE, "%s/%s", DATADIR, REPOS[r].name);
-	xmkdir (path, 0777);
+	snprintf (path, BUFFSIZE, "%s/%s", cache, REPOS[r].name);
+	rmkdir (path, 0777);
 	
 	if (!REPOS[r].SLACKWARE)
 	 continue; /* extra repository haven't patches and extra branches */
 	
 	/* repository patches directory */
-	snprintf (path, BUFFSIZE, "%s/%s/%s", DATADIR, REPOS[r].name, dbranch[1]);
-	xmkdir (path, 0777);
+	snprintf (path, BUFFSIZE, "%s/%s/%s", cache, REPOS[r].name, dbranch[1]);
+	rmkdir (path, 0777);
 	/* repository extra directory */
-	snprintf (path, BUFFSIZE, "%s/%s/%s", DATADIR, REPOS[r].name, dbranch[2]);
-	xmkdir (path, 0777);
+	snprintf (path, BUFFSIZE, "%s/%s/%s", cache, REPOS[r].name, dbranch[2]);
+	rmkdir (path, 0777);
 	
 	if (REPOS[r].SLACKWARE)
 	 slackware = true;
@@ -228,57 +212,83 @@ update_dir (void)
 	return;
 }
 
-/* retrieve a file list of REPOS[repository_index] */
+/* retrieve a file list of REPOS[index] */
 static int
-update_get_file (unsigned repository_index, branches branch, const char *file)
+update_get_file (unsigned index, branches branch, const char *file)
 {
-	unsigned i = repository_index;
-	char src [BUFFSIZE0] = { 0 },
-		 dest[BUFFSIZE0] = { 0 },
-		 msg [BUFFSIZE0] = { 0 };
-    bool timestamp = !opt.force;
+	net_t netpolicy;
+    const char *cache   = opt.cache;
+    const char *path    = REPOS[index].path;
+    const char *name    = REPOS[index].name;
+	const char *host    = REPOS[index].hostname;
+    const unsigned port = REPOS[index].port;
+    const char *act     = (REPOS[index].proto_t == proto_file) ? "Copying" : "Downloading";
+    const char *dfile   = file;
+    const bool official = REPOS[index].SLACKWARE;
+    const bool manifest = (bool) !strcmp (file, ZMANIFEST);
+
+    if (official && branch == _PACKAGES && manifest)
+    {
+        /* Fix MANIFEST.bz2 for slackware stable packages branch
+         * changing 'MANIFEST.bz2" to "slackware/MANIFEST.bz2".
+         * We preserve original file using `dfile' just for
+         * file destination path.
+         * In other case `file' and `dfile' will be same string !
+         */
+         file  = SMANIFEST;
+         dfile = ZMANIFEST;
+     }
+    
+    init_net_t (&netpolicy, REPOS[index].proto_t, port);
+
+    netpolicy.savepart   = true;
+    netpolicy.overwrite  = true;
+    netpolicy.checkstamp = !opt.force;
+    strncpy (netpolicy.hostname, host, BUFFSIZE);
 
     /* if we are downloading checksums list or manifest, file list or packages list was changed.
      * In this case we can skip timestamp check, otherwise file (checksum list or manifest) will
      * seem to be always update !
      */
-    if (!strcmp (file, CHECKSUMS) ||
-        !strcmp (file, ZMANIFEST))
-    timestamp = false;
-		
+    if (!strcmp (dfile, CHECKSUMS)
+     || !strcmp (dfile, ZMANIFEST))
+    {
+        netpolicy.checkstamp = false;
+    }
+
+
 	/* make (source and destination) path of our file and a message */
-	switch (branch) {
-		case _PACKAGES:
-			if (REPOS[i].SLACKWARE && strcmp (file, ZMANIFEST) == 0)
-			 snprintf (src,  BUFFSIZE, "%s/slackware/%s", REPOS[i].path, file);
-            else
-             snprintf (src,  BUFFSIZE, "%s/%s", REPOS[i].path, file);
-            snprintf (dest, BUFFSIZE, "%s/%s/%s", DATADIR, REPOS[i].name, file);
-			snprintf (msg,  BUFFSIZE, " `- Downloading %s ", file);
-			break;
-		case _PATCHES:
-			snprintf (src,  BUFFSIZE, "%s/patches/%s", REPOS[i].path, file);
-			snprintf (dest, BUFFSIZE, "%s/%s/patches/%s", DATADIR, REPOS[i].name, file);
-			snprintf (msg,  BUFFSIZE, " `- Downloading patches %s ", file);
-			break;
-		case _EXTRA:
-			snprintf (src,  BUFFSIZE, "%s/extra/%s", REPOS[i].path, file);
-			snprintf (dest, BUFFSIZE, "%s/%s/extra/%s", DATADIR, REPOS[i].name, file);
-			snprintf (msg,  BUFFSIZE, " `- Downloading extra %s ", file);
-			break;
-		default:
-			slassert (NULL);
-			break;
+	switch (branch)
+    {
+    case _PACKAGES:
+        snprintf (netpolicy.srcpath, BUFFSIZE, "%s/%s", path, file);
+        snprintf (netpolicy.destpath, BUFFSIZE, "%s/%s/%s", cache, name, dfile);
+        snprintf (netpolicy.msg, BUFFSIZE, "`- %s %s", act, dfile);
+    break;
+    case _PATCHES:
+        snprintf (netpolicy.srcpath,  BUFFSIZE, "%s/patches/%s", path, file);
+        snprintf (netpolicy.destpath, BUFFSIZE, "%s/%s/patches/%s", cache, name, dfile);
+        snprintf (netpolicy.msg,      BUFFSIZE, "`- %s patches %s", act, dfile);
+    break;
+    case _EXTRA:
+        snprintf (netpolicy.srcpath,  BUFFSIZE, "%s/extra/%s", path, file);
+        snprintf (netpolicy.destpath, BUFFSIZE, "%s/%s/extra/%s", cache, name, dfile);
+        snprintf (netpolicy.msg,      BUFFSIZE, "`- %s extra %s", act, dfile);
+    break;
+    default:
+        slassert (NULL);
+        break;
 	
 	}
 		
-	if (REPOS[i].proto_t == http)
-	 return get_file_over_http (msg, REPOS[i].hostname, src, dest, true, true, timestamp);
-	else
-	if (REPOS[i].proto_t == ftp)
-	 return get_file_over_ftp  (msg, REPOS[i].hostname, src, dest, true, true, timestamp);
+    switch (REPOS[index].proto_t)
+    {
+        case proto_http: return get_file_over_http (&netpolicy);  break;
+        case proto_ftp:  return get_file_over_ftp  (&netpolicy);  break;
+        case proto_file: return get_file_over_fs   (&netpolicy);  break;
+        default:   slassert (NULL);                               break;
+    }
 
-	slassert (NULL); /* error */
 	return 0;
 }
 
@@ -288,7 +298,7 @@ explode_manifest (unsigned i, branches b)
 {
     const char *sbranches[BRANCHES] = { "patches", "\b", "extra" };
     
-    fprintf (stdout, " >> Uncompressing %s %s: ", sbranches[b], ZMANIFEST);
+    fprintf (stdout, ">> Uncompressing %s %s: ", sbranches[b], ZMANIFEST);
     fflush (stdout);
 
 	if (!bz_decompress (make_rpath (i, b, ZMANIFEST), NULL))
@@ -304,49 +314,88 @@ down_ok (int rval)
     return (rval == FILE_DOWNLOADED || rval == FILE_ALREADY_UPDATE);
 }
 
+static void
+update_mirror_index (unsigned index)
+{
+    const char *cache = opt.cache;
+    FILE *fd;
+    char ipath[BUFFSIZE0];
+
+    snprintf (ipath, BUFFSIZE, "%s/slackware/mirror.slackyd", cache);
+    
+    /* update index of used repository */
+    if ((fd = fopen (ipath, "w")))
+    {
+        fprintf (fd, "[%s][%s]\n", REPOS[index].hostname, REPOS[index].path);
+        fclose (fd);
+    }
+    else
+    {
+        printfw ("\nWarning: cannot save used mirror.\n"
+                 "%s: %s\n\n", ipath, strerror (errno));
+    }
+
+    return;
+}
+
 /* update repository file list and checksum list */
 int
 update (void)
 {
-    #define NFILES 3
-
+    const char *cache = opt.cache;
     branches official_branches[BRANCHES] = { _PACKAGES, _PATCHES, _EXTRA };
-    const char *official_files[NFILES]    = { PACKAGES, CHECKSUMS, ZMANIFEST };
-    unsigned r = 0, fdown = 0, fall = 10, i, j;
-    bool slackware = false, tmpopt;
+    const char *official_files[] = { PACKAGES, ZMANIFEST, NULL };
+    const unsigned fall = 8;
+    unsigned r = 0, fdown = 0, i, j;
+    bool slackware = false;
     int rval;
-    
-    tmpopt = opt.force;
-    if (packages_list_check (false, false) > 0)
+    char ipath[BUFFSIZE0];
+        
+    if (!opt.force && opt.warn && (rval = packages_list_check (false, PKGLIST_CHECK_QUIT_NEVER)) > 0)
     {
-        if (opt.warn)
-         fprintf (stderr, "*** Warning: broken packages list, forcing update.\n\n");
-        opt.force = true;
+        fprintf (stderr,
+         "*** Warning: Broken packages list, %d file%s missing.\n"
+         "*** Try to force update or fix your slackyd.conf.\n\n",
+         rval, (rval > 1) ? "s are" : " is");
     }
 
     update_dir ();
-
+    snprintf (ipath, BUFFSIZE, "%s/slackware/mirror.slackyd", cache);
+    
     /* Update official slackware files list */
-    while (r < REPOS[0].N && REPOS[r].SLACKWARE && !slackware && fdown != fall)
+    while (r < N_REPOS && REPOS[r].SLACKWARE && !slackware && fdown != fall)
     {
-        fprintf (stdout, "Repository: %s [%s]\n", REPOS[r].name, REPOS[r].hostname);
-
+        fprintf (stdout, "Repository: %s [%s]", REPOS[r].name, REPOS[r].hostname);
+        if (REPOS[r].port) {
+            fprintf (stdout, "[port: %d]", REPOS[r].port);
+        }
+        putchar ('\n');
+        
         /* First of all update FILELIST.TXT;
-         * If it is already update, we assume that all is update. Else
+         * If it is already update, we assume that all is update and break. Else
          * if it cannot be downloaded we assume that repository is not standard.
+         * After download CHECKUMS.md5..."
          */
-        if ((rval  = update_get_file (r, _PACKAGES, FILELIST)) != FILE_DOWNLOADED)
+        rval = update_get_file (r, _PACKAGES, FILELIST);
+        if (rval == FILE_ALREADY_UPDATE)
         {
+            if (!file_exist (ipath, NULL, true))
+             update_mirror_index (r);
             break;
         }
+
+        if (down_ok (rval) &&
+            down_ok ((rval = update_get_file (r, _PACKAGES, CHECKSUMS))))
+        {
+            fdown = 2; /* FILELIST.TXT and CHECKSUMS.md5 was retreived. */
+        }
         
-        fdown = 1; /* FILELIST.TXT was retreived. */
-        
-        /* For each branch, /, /patches and /extra ... */
+
+       /* For each branch, /, /patches and /extra ... */
         for (i = 0; down_ok (rval) && i < BRANCHES; i++)
         {
             /* For each file, PACKAGES.TXT, CHECKSUMS.TXT and MANIFEST.bz2 ... */
-            for (j = 0; down_ok (rval) && j < NFILES; j++)
+            for (j = 0; down_ok (rval) && official_files[j]; j++)
             {
                 rval = update_get_file (r, official_branches[i], official_files[j]);
 
@@ -357,7 +406,7 @@ update (void)
                 if (!strcmp (official_files[j], PACKAGES)
                  && rval == FILE_ALREADY_UPDATE)
                 {
-                    fdown += 3;
+                    fdown += 2; /* packages.txt and manifest.bz2 */
                     break;
                 }
 
@@ -372,19 +421,26 @@ update (void)
             }
         }
 
-        r++;
         slackware = true;
-        
+       
         if (fdown != fall)
         {
+            remove (ipath);
             fprintf (stderr, "\nWarning: cannot retreive all file (%u/%u). Trying next: %s\n\n",
-                fdown, fall, (r < REPOS[0].N && REPOS[r].SLACKWARE) ? "found" : "nothing");
+                fdown, fall, (r + 1 < N_REPOS && REPOS[r + 1].SLACKWARE) ? "found" : "nothing");
             slackware = false;
         }
+        else
+        {
+            update_mirror_index (r);
+        }
+
+        r++;
+        
     }
 
     /* Update extra repository files list */
-    while (r < REPOS[0].N)
+    while (r < N_REPOS)
     {
         if (REPOS[r].SLACKWARE)
         {
@@ -392,8 +448,12 @@ update (void)
             continue;
         }
          
-        fprintf (stdout, "Repository: %s [%s]\n", REPOS[r].name, REPOS[r].hostname);
-
+        fprintf (stdout, "Repository: %s [%s]", REPOS[r].name, REPOS[r].hostname);
+        if (REPOS[r].port) {
+            fprintf (stdout, "[port: %d]", REPOS[r].port);
+        }
+        putchar ('\n');
+        
         if (update_get_file (r, _PACKAGES, FILELIST)  == FILE_DOWNLOADED
          && update_get_file (r, _PACKAGES, PACKAGES)  == FILE_DOWNLOADED
          && update_get_file (r, _PACKAGES, CHECKSUMS) == FILE_DOWNLOADED
@@ -405,10 +465,7 @@ update (void)
         r++;
     }
         
-    opt.force = tmpopt;
-        
-  
-	return EXIT_SUCCESS;
+    return EXIT_SUCCESS;
 }
 
 	
@@ -431,70 +488,64 @@ pkg_u_installed (const char *pkg)
 static int
 search_new (const char *pkgname, upgrade_t *toupgrade[])
 {
-	unsigned i, j, pnew = 0, sub_new = 0, checked = 0;
-    int rval = 0, diff[2], sflags, cflags;
+	unsigned i, j, pnew = 0, sub_new = 0;
+    int diff[2], sflags, cflags;
     regex_t preg;
     pkg_t **packages = NULL;
     unsigned npackages = 0x00;
         
-       
     fprintf (stdout, "Reading packages data: ");
     fflush (stdout);
 
     sflags = (pkgname && opt.use_regex) ? MATCH_USING_REGEX : MATCH_EXACTLY_NAME_INST;
-    rval = search (sflags, pkgname, &packages, &npackages);  
-    if (rval == EXIT_FAILURE) {
+
+    packages = pkgl2pkgt (search (sflags, pkgname), &npackages);  
+
+    if (!npackages)
+    {
          if (!pkgname)
           fprintf (stdout, "\n\n"
           "No packages found !?\n"
           "Try to update packages list.\n");
          else
           fprintf (stdout, "nothing to do for `%s'.\n", pkgname);
-          
          return EXIT_FAILURE;
     }
  
     fprintf (stdout, "done.\n");
-
-    if (!opt.verbose) {
-        fprintf (stdout, "Searching new packages: ");
-        fflush (stdout);
-    }
-    else
-    putchar ('\n');
-    
-    
+        
     cflags = REG_EXTENDED | REG_NOSUB;
-    if (opt.use_regex && pkgname) {
+    if (opt.use_regex && pkgname)
+    {
         if (opt.case_insensitive)
          cflags |= REG_ICASE;
         xregcomp (&preg, pkgname, cflags);
     }
 
-    for (i = 0; i < nginst; i++) {
+    for (i = 0; i < nginst; i++)
+    {
 
-    if ((pkgname && opt.use_regex == true  && regexec (&preg, ginst[i].name, 0, NULL, 0))
-     || (pkgname && opt.use_regex == false && !xstrstr (ginst[i].name, pkgname)))
+    if ((pkgname && opt.use_regex  && regexec (&preg, ginst[i].name, 0, NULL, 0))
+     || (pkgname && !opt.use_regex && !xstrstr (ginst[i].name, pkgname)))
      {
          continue;
      }
      
-    if (opt.verbose) {
-        fprintf (stdout, "Verifying %s: ", ginst[i].name);
-        fflush (stdout);
-    }
+    printfv ("Verifying %s: ", ginst[i].name);
     
-    checked++;
-    sub_new  = 0;
+    sub_new = 0;
     
-    for (j = 0; j < npackages; j++) {
+    for (j = 0; j < npackages; j++)
+    {
 
         /* if current available package haven't same name of current
          * package installed, of if already installed, go to begin for()
          */
         if (strcmp (packages[j]->pstruct.name, ginst[i].pstruct.name)
-        ||  pkg_u_installed (packages[j]->name))
-         continue;
+          || pkg_u_installed (packages[j]->name))
+        {
+            continue;
+        }
 
         diff[0] = strverscmp (ginst[i].pstruct.version, packages[j]->pstruct.version);
 	    diff[1] = atoi (ginst[i].pstruct.build) - atoi (packages[j]->pstruct.build);
@@ -512,41 +563,28 @@ search_new (const char *pkgname, upgrade_t *toupgrade[])
                 (*toupgrade)[pnew].remote = NULL;
                 (*toupgrade)[pnew].local  = xstrdup (ginst[i].name);
             }
-            add_pkg_t (&(*toupgrade)[pnew].remote, &sub_new, packages[j]);
+            pkg_t_add (&(*toupgrade)[pnew].remote, &sub_new, packages[j]);
         }
     } /* end available packages for() */
     
-    if (sub_new > 0) {
+    if (sub_new > 0)
+    {
        	(*toupgrade)[pnew].sub_new = sub_new;
        	pnew++;
-       	if (opt.verbose)
-       	{
-       	    fprintf (stdout, "found %u new.\n", sub_new);
-        }
+       	printfv ("found %u new.\n", sub_new);
     }
     else
-    if (opt.verbose)
     {
-    	fprintf (stdout, "nothing to do.\n");
+        printfv ("nothing to do.\n");
     }
-
+    
 	} /* end installed packages list for() */
 
     
     if (opt.use_regex && pkgname)
      regfree (&preg);
-    free_pkgs_t (&packages, npackages);
+    pkg_t_free_all (&packages, npackages);
     
-    if (!opt.verbose)
-     fprintf (stdout, "done !\n");
-
-    
-    if (checked == 0) {
-        fprintf (stdout, "\n"
-        "There aren't packages \"%s\" installed.\n", pkgname);
-        return EXIT_FAILURE;
-    }
-        
     putchar ('\n');
 
     if (pnew == 0)
@@ -561,7 +599,7 @@ search_new (const char *pkgname, upgrade_t *toupgrade[])
         }
     }
 
-	return (pnew > 0) ? (int) pnew : EXIT_FAILURE;
+	return (pnew > 0) ? (int) pnew : -1;
 }
 
 
@@ -569,27 +607,31 @@ search_new (const char *pkgname, upgrade_t *toupgrade[])
 int
 upgrade (const char *pkgname)
 {
-    pkg_t **gpkg = NULL;
+    pkg_t **gpkg = NULL, **rptr;
     int i, pnew, iflags;
     unsigned j, ngpkg = 0;
     upgrade_t *toupgrade = NULL;
     char c = 0;
     
-    pkg_t **prequired = NULL;
+    choose_dep_t ddata;
+	pkg_t **prequired = NULL;
     unsigned nprequired = 0;
     bool ch;
 
-    packages_list_check (true, true);
+    memset (&ddata, 0, sizeof (ddata));
+
+	packages_list_check (true, PKGLIST_CHECK_QUIT_ON_ALL);
 
     iflags = (pkgname && opt.use_regex) ? MATCH_USING_REGEX : MATCH_ALL_NAME;
-    if (pkgname && is_installed (pkgname, iflags, 0) == PACKAGE_STATUS_NOT_INSTALLED) {
-        fprintf (stderr,
-        "No packages \"%s\" are installed.\n", pkgname);
+    if (pkgname && is_installed (pkgname, iflags, 0) == PACKAGE_STATUS_NOT_INSTALLED)
+    {
+        fprintf (stderr, "No packages \"%s\" are installed.\n", pkgname);
         return EXIT_FAILURE;
     }
     
 
-    if ((pnew = search_new (pkgname, &toupgrade)) < 1) {
+    if ((pnew = search_new (pkgname, &toupgrade)) < 1)
+    {
         return EXIT_FAILURE;
     }
 
@@ -597,111 +639,137 @@ upgrade (const char *pkgname)
     "Found %d package%s to upgrade:\n"
     "----------------------------- \n", pnew, (pnew > 1) ? "s" : "");
 
-    for (i = 0; i < pnew; i++) {
-        printf ("%s:\n", toupgrade[i].local);
-        fflush (stdout);
-
-        for (j = 0; j < toupgrade[i].sub_new - 1; j++) {
-            printf ("\t%s %s\n",
-            toupgrade[i].remote[j]->name, pkg_type (toupgrade[i].remote[j]));
+    for (i = 0; i < pnew; i++)
+    {
+        rptr = toupgrade[i].remote;
+        fprintf (stdout, "%s:\n", toupgrade[i].local);
+        for (j = 0; j < toupgrade[i].sub_new - 1; j++)
+        {
+            fprintf (stdout, "\t%s %s\n", rptr[j]->name, pkg_type (rptr[j]));
         }
-        printf ("\t%s %s\n",
-        toupgrade[i].remote[j]->name, pkg_type (toupgrade[i].remote[j]));    
+        fprintf (stdout, "\t%s %s\n", rptr[j]->name, pkg_type (rptr[j]));    
 	}
 
     putchar ('\n');
 
     /* choice package to upgrade */
-    for (i = 0; i < pnew; i++) {
+    for (i = 0; i < pnew; i++)
+    {
 
+    rptr = toupgrade[i].remote;
     /* set `ch' to false; we use this variable to track package choice */
     ch = false;
 
-    sort_pkg (&toupgrade[i].remote, &toupgrade[i].sub_new);
+    if (toupgrade[i].sub_new > 1)
+    {
+        sort_pkg (&rptr, &toupgrade[i].sub_new);
 
-    if (toupgrade[i].sub_new > 1) {
-        fprintf (stdout, "Found %u different new version of package %s:\n",
-        toupgrade[i].sub_new, toupgrade[i].remote[0]->pstruct.name);
-        for (j = 0; j < toupgrade[i].sub_new; j++)
-         fprintf (stdout, "  --> %s %s\n",
-         toupgrade[i].remote[j]->name, pkg_type (toupgrade[i].remote[j]));
+		fprintf (stdout, "Found %u different new version of package %s:\n",
+            toupgrade[i].sub_new,
+            rptr[0]->pstruct.name);
+
+		for (j = 0; j < toupgrade[i].sub_new; j++)
+        {
+            fprintf (stdout, "  --> %s %s\n", rptr[j]->name, pkg_type (rptr[j]));
+        }
     }
 
-    for (j = 0; j < toupgrade[i].sub_new; j++) {
+    for (j = 0; j < toupgrade[i].sub_new; j++)
+    {
 
-        if (!opt.force) {
-            if (j == 0)
-             fprintf (stdout, "Upgrade %s to:\n", toupgrade[i].local);
+        /* Check for package already in dependencies list */
+        if (pkg_t_check (rptr[j]->pstruct.name, prequired, nprequired, MATCH_EXACTLY_NAME))
+        {
+            break;
+        }
 
-            fprintf (stdout, "\t%s %s ? [y/N] ", toupgrade[i].remote[j]->name,
-            (toupgrade[i].sub_new == 1) ? pkg_type (toupgrade[i].remote[j]) : "\b");
+        if (!opt.force)
+        {
+            if (j == 0) {
+                fprintf (stdout, "Upgrade %s to:\n", toupgrade[i].local);
+            }
+            fprintf (stdout, "\t%s %s ? [y/N] ",
+                rptr[j]->name, (toupgrade[i].sub_new == 1) ? pkg_type (rptr[j]) : "\b");
         }
         else
-         if (toupgrade[i].sub_new > 1)
-         fprintf (stdout, "Will be used %s !\n", toupgrade[i].remote[0]->name);
+        if (toupgrade[i].sub_new > 1) {
+            fprintf (stdout, "Will be used %s !\n", rptr[0]->name);
+        }
 
         fflush (stdout);
 
-        if (opt.force == true || (c = getchar()) == 'Y' || c == 'y') {
+        if (opt.force || (c = getchar()) == 'Y' || c == 'y')
+        {
 
             ch = true;
             ngpkg++;
 
             gpkg = (pkg_t **)
                 xrealloc (gpkg, ngpkg * sizeof (pkg_t *));
-            gpkg[ngpkg - 1] = (pkg_t *)
-                xmalloc (sizeof (pkg_t));
-
-            cp_pkg (toupgrade[i].remote[j], &gpkg[ngpkg - 1]);
+            gpkg[ngpkg - 1] = pkg_t_dup (rptr[j]);
 
             if (!opt.force)
-             clrbuff();
+            {
+                clrbuff();
+            }
+            
             break;
         }
-        if (c != '\n')
-         clrbuff();
+        
+        if (c != '\n') {
+            clrbuff();
         }
+    }
 
-        free (toupgrade[i].local);
-        free_pkgs_t (&toupgrade[i].remote, toupgrade[i].sub_new);
+    free (toupgrade[i].local);
+    pkg_t_free_all (&toupgrade[i].remote, toupgrade[i].sub_new);
 
-        if (ch && opt.check_dep) {
-
-            search_mreq (&gpkg[ngpkg - 1], 0);
-            
-	        /* search missing packages dependencies in repository. If found ask if download. */
-	        if (opt.check_dep && opt.resolve_dep && gpkg[ngpkg - 1]->nmrequired > 0)
-	        {
-	            choice_dep (&prequired, &nprequired, gpkg[ngpkg - 1], gpkg, ngpkg);
-            }
+    if (ch && opt.check_dep)
+    {
+        /* search missing packages dependencies in repository. */
+        if (opt.check_dep && opt.resolve_dep)
+        {
+			ddata.next  = gpkg;
+			ddata.nnext = ngpkg;
+            choice_dep (&ddata, gpkg[ngpkg - 1], DEP_T_REQUIRED);
+			choice_dep (&ddata, gpkg[ngpkg - 1], DEP_T_SUGGESTED);
+		}
 	}
 	}
 
-	/* Free local missing packages list */
-    choice_dep (NULL, NULL, NULL, 0, 0);
-	
-    free (toupgrade);
+	free (toupgrade);
 
     if (!opt.force)
      putchar ('\n');
     
-    /* download packages and packages required */
-    for (j = 0; j < nprequired; j++) {
-        if (get_pkg (prequired[j]) == EXIT_SUCCESS)
-         verify_md5 (prequired[j]);
-     }
+    /* download choosed packages suggested and required */
+    if ((prequired = pkgl2pkgt (ddata.added, &nprequired)))
+	{
+		for (j = 0; j < nprequired; j++)
+    	{
+        	if (get_pkg (prequired[j]) == EXIT_SUCCESS)
+        	{
+            	verify_md5 (prequired[j]);
+        	}
+		}
+		pkg_t_free_all (&prequired, nprequired);
+    }
   
-    free_pkgs_t (&prequired, nprequired);
     
     if (ngpkg == 0)
-      fprintf (stdout, "Nothing was choiced.\n");
+     fprintf (stdout, "Nothing was choosed.\n");
     else
-      for (j = 0; j < ngpkg; j++) {
-        if (get_pkg (gpkg[j]) == EXIT_SUCCESS)
-         verify_md5 (gpkg[j]);
-       }
-
-    free_pkgs_t (&gpkg, ngpkg);
+    {
+        for (j = 0; j < ngpkg; j++)
+        {
+            if (get_pkg (gpkg[j]) == EXIT_SUCCESS)
+            {
+                verify_md5 (gpkg[j]);
+            }
+        }
+    }
+    
+    pkg_t_free_all (&gpkg, ngpkg);
     
     return EXIT_SUCCESS;
 
